import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.Serializable;
import java.io.StringReader;
import java.io.Writer;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Random;
import java.util.Vector;

import marf.MARF;
import marf.Classification.Distance.HammingDistance;
import marf.Classification.Similarity.CosineSimilarityMeasure;
import marf.Classification.Stochastic.MaxProbabilityClassifier;
import marf.Classification.Stochastic.ZipfLaw;
import marf.Stats.StatisticalEstimators.GLI;
import marf.Stats.StatisticalEstimators.KatzBackoff;
import marf.Stats.StatisticalEstimators.MLE;
import marf.Stats.StatisticalEstimators.SLI;
import marf.Stats.StatisticalEstimators.StatisticalEstimator;
import marf.Stats.StatisticalEstimators.Smoothing.AddDelta;
import marf.Stats.StatisticalEstimators.Smoothing.AddOne;
import marf.Stats.StatisticalEstimators.Smoothing.GoodTuring;
import marf.Stats.StatisticalEstimators.Smoothing.WittenBell;
import marf.Storage.ModuleParams;
import marf.Storage.Result;
import marf.Storage.ResultSet;
import marf.Storage.TrainingSet;
import marf.nlp.util.NLPStreamTokenizer;
import marf.util.Debug;
import marf.util.MARFException;
import marf.util.MARFRuntimeException;
import marf.util.NotImplementedException;
import marf.util.OptionProcessor;
import SATE.Report;
import SATE.SATEDb;
import SATE.Warning;
import SATE.heuristics.lines.FirstLineNumber;
import SATE.heuristics.lines.ILineNumberHeuristic;
import SATE.heuristics.lines.MidLineNumber;
import Storage.AnyToWAVLoader;
import Storage.FileItem;
import Storage.GenericMetaXMLLoader;
import Storage.LineContextReader;
import Storage.MetaLocation;
import Storage.RawSampleLoader;

/**
 * <p>
 * Application for the SATE2010, based on the MARF framework, <a
 * href="http://marf.sf.net">http://marf.sf.net</a>.
 * 
 * Details are documented in <a href="http://arxiv.org/abs/1010.2511">http://arxiv.org/abs/1010.2511</a>
 * </p>
 * 
 * @author Serguei Mokhov
 * 
 * @version $Id: MARFCATApp.java,v 1.33 2011/01/14 00:31:30 mokhov Exp $
 * @since July 2010, MARF's v. 0.3.0.6
 */
public class MARFCATApp
{
	/*
	 * ---------------- Apps. Versioning ----------------
	 */

	/**
	 * Current major version of the application.
	 */
	public static final int MAJOR_VERSION = 0;

	/**
	 * Current minor version of the application.
	 */
	public static final int MINOR_VERSION = 0;

	/**
	 * Current revision of the application.
	 */
	public static final int REVISION = 1;

	/*
	 * ----------------------------------
	 * Defaults
	 * ----------------------------------
	 */

	/**
	 * Default file name extension for sample files; case-insensitive.
	 */
	public static final String DEFAULT_SAMPLE_FILE_EXTENSION = "xml";

	/*
	 * ----------------------------------
	 * Major and Misc Options Enumeration
	 * ----------------------------------
	 */

	/**
	 * Numeric equivalent of the option <code>--train</code>.
	 */
	public static final int OPT_TRAIN = 0;

	/**
	 * Numeric equivalent of the option <code>--ident</code>.
	 */
	public static final int OPT_IDENT = 1;

	/**
	 * Numeric equivalent of the option <code>--stats</code>.
	 */
	public static final int OPT_STATS = 2;

	/**
	 * Numeric equivalent of the option <code>--reset</code>.
	 */
	public static final int OPT_RESET = 3;

	/**
	 * Numeric equivalent of the option <code>--version</code>.
	 */
	public static final int OPT_VERSION = 4;

	/**
	 * Numeric equivalent of the option <code>--help</code>.
	 */
	public static final int OPT_HELP_LONG = 5;

	/**
	 * Numeric equivalent of the option <code>-h</code>.
	 */
	public static final int OPT_HELP_SHORT = 6;

	/**
	 * Numeric equivalent of the option <code>-debug</code>.
	 */
	public static final int OPT_DEBUG = 7;

	/**
	 * Numeric equivalent of the option <code>-graph</code>.
	 */
	public static final int OPT_GRAPH = 8;

	/**
	 * Numeric equivalent of the option <code>-spectrogram</code>.
	 */
	public static final int OPT_SPECTROGRAM = 9;

	/**
	 * Numeric equivalent of the option <code>&lt;writer ID&gt;</code>.
	 */
	public static final int OPT_EXPECTED_ID = 10;

	/**
	 * Numeric equivalent of the option <code>--batch-ident</code>.
	 */
	public static final int OPT_BATCH_IDENT = 11;

	/**
	 * Numeric equivalent of the option <code>--single-train</code>.
	 */
	public static final int OPT_SINGLE_TRAIN = 12;

	/**
	 * Numeric equivalent of the option
	 * <code>&lt;sample-file-or-directory-name&gt;</code>.
	 */
	public static final int OPT_DIR_OR_FILE = 13;

	/**
	 * Numeric equivalent of the option <code>--best-score</code>.
	 */
	public static final int OPT_BEST_SCORE = 14;

	/**
	 * Numeric equivalent of the option <code>--gui</code> to have a GUI.
	 */
	public static final int OPT_GUI = 15;

	/**
	 * Numeric equivalent of the option <code>-noise</code> to enable noise
	 * removal.
	 */
	public static final int OPT_NOISE_REMOVAL = 16;

	/**
	 * Numeric equivalent of the option <code>-silence</code> to enable silence
	 * removal.
	 */
	public static final int OPT_SILENCE_REMOVAL = 17;

	/**
	 * Numeric equivalent of the option <code>--show-training-set</code> to
	 * display text representation of a training set file.
	 */
	public static final int OPT_SHOW_TRAINING_SET_DATA = 18;

	// SATE-specific

	public static final int OPT_SATE_SAMPLE_FILE = 19;

	public static final int OPT_NO_PREPREP = 20;

	/**
	 * -cweid
	 */
	public static final int OPT_SATE_CWEID = 26;

	public static final int OPT_SILENCE_NOISE_REMOVAL = 27;

	public static final int OPT_SATE_TEST_REFERENCE_DATA = 28;

	public static final int OPT_TRAINING_MODEL_MEAN_CLUSTER = 29;
	public static final int OPT_TRAINING_MODEL_MEDIAN_CLUSTER = 30;
	public static final int OPT_TRAINING_MODEL_FEATURE_SET = 31;

	/*
	 * ---------------------
	 * State Data Structures
	 * ---------------------
	 */

	/**
	 * Instance of the database of writers to track stats per configuration for
	 * all writers.
	 */
	protected static SATEDb soDBPerConfig = null;

	/**
	 * Instance of the database of writers to track stats per writer for all
	 * configuration.
	 */
	protected static SATEDb soDBPerCVE = null;

	protected static SATEDb soDBPerCWE = null;

	protected static GenericMetaXMLLoader soRefDataLoader = GenericMetaXMLLoader.getInstance();

	/**
	 * Instance of the option processing utility.
	 */
	protected static OptionProcessor soGetOpt = new OptionProcessor();

	/**
	 * Stream of incoming tokens.
	 */
	private static NLPStreamTokenizer soTokenizer = null;

	/**
	 * Reader object for incoming text files.
	 */
	private static BufferedReader soReader;

	/*
	 * ----------------- Static State Init -----------------
	 */

	static
	{
		// Main options
		soGetOpt.addValidOption(OPT_TRAIN, "--train");
		soGetOpt.addValidOption(OPT_SINGLE_TRAIN, "--single-train");
		soGetOpt.addValidOption(OPT_IDENT, "--ident");
		soGetOpt.addValidOption(OPT_BATCH_IDENT, "--batch-ident");
		soGetOpt.addValidOption(OPT_STATS, "--stats");
		soGetOpt.addValidOption(OPT_BEST_SCORE, "--best-score");
		soGetOpt.addValidOption(OPT_SHOW_TRAINING_SET_DATA, "--show-training-set");

		soGetOpt.addValidOption(OPT_VERSION, "--version");
		soGetOpt.addValidOption(OPT_RESET, "--reset");
		soGetOpt.addValidOption(OPT_HELP_LONG, "--help");
		soGetOpt.addValidOption(OPT_HELP_SHORT, "-h");
		soGetOpt.addValidOption(OPT_DEBUG, "--debug");

		// Sample Loading
		soGetOpt.addValidOption(MARF.WAV, "-wav");
		soGetOpt.addValidOption(MARF.TEXT, "-text");
		soGetOpt.addValidOption(MARF.TEXT + 1, "-tiff");
		soGetOpt.addValidOption(MARF.TEXT + 2, "-rawbyte");

		// Preprocessing
		soGetOpt.addValidOption(OPT_NOISE_REMOVAL, "-noise");
		soGetOpt.addValidOption(OPT_SILENCE_REMOVAL, "-silence");
		soGetOpt.addValidOption(OPT_SILENCE_NOISE_REMOVAL, "-silence-noise");

		soGetOpt.addValidOption(MARF.DUMMY, "-norm");
		soGetOpt.addValidOption(MARF.HIGH_FREQUENCY_BOOST_FFT_FILTER, "-boost");
		soGetOpt.addValidOption(MARF.HIGH_PASS_FFT_FILTER, "-high");
		soGetOpt.addValidOption(MARF.LOW_PASS_FFT_FILTER, "-low");
		soGetOpt.addValidOption(MARF.BANDPASS_FFT_FILTER, "-band");
		soGetOpt.addValidOption(MARF.BAND_STOP_FFT_FILTER, "-bandstop");
		soGetOpt.addValidOption(MARF.HIGH_PASS_BOOST_FILTER, "-highpassboost");
		soGetOpt.addValidOption(MARF.RAW, "-raw");
		soGetOpt.addValidOption(MARF.ENDPOINT, "-endp");
		soGetOpt.addValidOption(MARF.HIGH_PASS_CFE_FILTER, "-highcfe");
		soGetOpt.addValidOption(MARF.LOW_PASS_CFE_FILTER, "-lowcfe");
		soGetOpt.addValidOption(MARF.BAND_PASS_CFE_FILTER, "-bandcfe");
		soGetOpt.addValidOption(MARF.BAND_STOP_CFE_FILTER, "-bandstopcfe");

		// Feature extraction
		soGetOpt.addValidOption(MARF.FFT, "-fft");
		soGetOpt.addValidOption(MARF.LPC, "-lpc");
		soGetOpt.addValidOption(MARF.RANDOM_FEATURE_EXTRACTION, "-randfe");
		soGetOpt.addValidOption(MARF.MIN_MAX_AMPLITUDES, "-minmax");
		soGetOpt.addValidOption(MARF.FEATURE_EXTRACTION_AGGREGATOR, "-aggr");
		soGetOpt.addValidOption(MARF.F0, "-f0");
		soGetOpt.addValidOption(MARF.SEGMENTATION, "-segm");
		soGetOpt.addValidOption(MARF.CEPSTRAL, "-cepstral");

		// Classification
		soGetOpt.addValidOption(MARF.NEURAL_NETWORK, "-nn");
		soGetOpt.addValidOption(MARF.EUCLIDEAN_DISTANCE, "-eucl");
		soGetOpt.addValidOption(MARF.CHEBYSHEV_DISTANCE, "-cheb");
		soGetOpt.addValidOption(MARF.MINKOWSKI_DISTANCE, "-mink");
		soGetOpt.addValidOption(MARF.MAHALANOBIS_DISTANCE, "-mah");
		soGetOpt.addValidOption(MARF.RANDOM_CLASSIFICATION, "-randcl");
		soGetOpt.addValidOption(MARF.DIFF_DISTANCE, "-diff");
		soGetOpt.addValidOption(MARF.ZIPFS_LAW, "-zipf");
		soGetOpt.addValidOption(MARF.MARKOV, "-markov");
		soGetOpt.addValidOption(MARF.HAMMING_DISTANCE, "-hamming");
		soGetOpt.addValidOption(MARF.COSINE_SIMILARITY_MEASURE, "-cos");

		// Graphical
		soGetOpt.addValidOption(OPT_SPECTROGRAM, "-spectrogram");
		soGetOpt.addValidOption(OPT_GRAPH, "-graph");
		soGetOpt.addValidOption(OPT_GUI, "--gui");

		// NLP
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.MLE, "-mle");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.ADD_ONE, "-add-one");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.ADD_DELTA, "-add-delta");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.ELE, "-ele");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.WITTEN_BELL, "-witten-bell");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.GOOD_TURING, "-good-turing");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.SLI, "-sli");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.GLI, "-gli");
		soGetOpt.addValidOption(MARF.EStatisticalEstimators.KATZ_BACKOFF, "-backoff");

		soGetOpt.addValidOption(MARF.NLP.STEMMING, "-stemming");

		soGetOpt.addValidOption(MARF.NLP.CHARACTER_MODE, "-char");
		soGetOpt.addValidOption(MARF.NLP.WORD_MODE, "-word");

		soGetOpt.addValidOption(MARF.NLP.CASE_SENSITIVE, "-case");
		soGetOpt.addValidOption(MARF.NLP.PARSE_NUMBERS, "-num");
		soGetOpt.addValidOption(MARF.NLP.PARSE_QUOTED_STRINGS, "-quote");
		soGetOpt.addValidOption(MARF.NLP.PARSE_ENDS_OF_SENTENCE, "-eos");
		soGetOpt.addValidOption(MARF.NLP.RAW_ZIPFS_LAW_DUMP, "-nolog");

		soGetOpt.addValidOption(MARF.ENgramModels.UNIGRAM, "-unigram");
		soGetOpt.addValidOption(MARF.ENgramModels.BIGRAM, "-bigram");
		soGetOpt.addValidOption(MARF.ENgramModels.TRIGRAM, "-trigram");

		soGetOpt.addValidOption(MARF.NLP.TRAIN, "--train-nlp");
		soGetOpt.addValidOption(MARF.NLP.CLASSIFY, "--ident-nlp");

		soGetOpt.addValidOption(MARF.NLP.ZIPFS_LAW_CHEAT, "--cheat");

		soGetOpt.addValidOption(MARF.NLP.INTERACTIVE, "-interactive");

		// SATE
		soGetOpt.addValidOption(OPT_NO_PREPREP, "-nopreprep");

		soGetOpt.addValidOption(OPT_SATE_CWEID, "-cweid");

		soGetOpt.addValidOption(OPT_SATE_TEST_REFERENCE_DATA, "-ref");
	}

	private static Report soReport = new Report();

	/**
	 * Main body.
	 * 
	 * @param argv
	 *            command-line arguments
	 */
	public static final void main(String[] argv)
	{
		try
		{
			// Since some new API is always introduced...
			validateVersions();

			// Parse extra arguments
			int iValidOptions = soGetOpt.parse(argv);

			// Debug.enableDebug();
			Debug.debug("Option set parsed: " + soGetOpt);

			if(iValidOptions == 0)
			{
				throw new Exception("No valid options found: " + soGetOpt);
			}

			// setCustomConfig();

			/*
			 * If supplied in the command line, the system when classifying,
			 * will output this ID next to the guessed one.
			 */
			int iExpectedID = -1;

			switch(soGetOpt.getInvalidOptions().size())
			{
				// Unknown
				case 0:
				{
					iExpectedID = -1;
					break;
				}

				case 3:
				{

					iExpectedID = Integer.parseInt(soGetOpt.getInvalidOptions().lastElement().toString());
					soGetOpt.addActiveOption(OPT_EXPECTED_ID, soGetOpt.getInvalidOptions().lastElement().toString());

					// no 'break' is intentional
				}

				/*
				 * Extract and make active and valid option out of a
				 * filename.
				 */
				case 2:
				{
					String strSampleFile = soGetOpt.getInvalidOptions().elementAt(1).toString();
					soGetOpt.addActiveOption(OPT_SATE_SAMPLE_FILE, strSampleFile);

					// No break required as the file or directory
					// must always be present. Also, the clearance
					// of the invalid options need to be postponed.
				}

				/*
				 * In the case of a single invalid option always assume it
				 * is either a filename or a directory name for the --ident
				 * or --train options.
				 */
				case 1:
				{
					soGetOpt.addActiveOption(OPT_DIR_OR_FILE, soGetOpt.getInvalidOptions().firstElement().toString());
					soGetOpt.getInvalidOptions().clear();
					break;
				}

				default:
				{
					throw new Exception("Unrecognized options found: " + soGetOpt.getInvalidOptions());
				}
			}

			/*
			 * Load the database
			 */
			if(soGetOpt.isActiveOption(OPT_SATE_SAMPLE_FILE))
			{
				soDBPerConfig = new SATEDb(soGetOpt.getOption(OPT_SATE_SAMPLE_FILE));
				soDBPerCVE = new SATEDb(soGetOpt.getOption(OPT_SATE_SAMPLE_FILE));
				soDBPerCWE = new SATEDb(soGetOpt.getOption(OPT_SATE_SAMPLE_FILE));
			}
			else
			{
				soDBPerConfig = new SATEDb(soGetOpt.getOption(OPT_DIR_OR_FILE));
				soDBPerCVE = new SATEDb(soGetOpt.getOption(OPT_DIR_OR_FILE));
				soDBPerCWE = new SATEDb(soGetOpt.getOption(OPT_DIR_OR_FILE));
			}

			if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
			{
				soDBPerConfig.setCWE(true);
				soDBPerCWE.setCWE(true);
			}

			soDBPerConfig.connect();
			soDBPerConfig.query();

			soDBPerCVE.connect();
			soDBPerCVE.query();

			soDBPerCWE.connect();
			soDBPerCWE.query();

			soDBPerConfig.setFilename("config." + soDBPerConfig.getFilename());
			soDBPerCVE.setFilename("cve." + soDBPerCVE.getFilename());
			soDBPerCWE.setFilename("cwe." + soDBPerCWE.getFilename());

			setDefaultConfig();
			setCustomConfig();

			// Don't mix up CVE and CWE training sets
			if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
			{
				MARF.setTrainingSetFilenamePrefix("cwe.");
			}
			else
			{
				MARF.setTrainingSetFilenamePrefix("cve.");
			}

			// Set misc configuration
			MARF.setDumpSpectrogram(soGetOpt.isActiveOption(OPT_SPECTROGRAM));
			MARF.setDumpWaveGraph(soGetOpt.isActiveOption(OPT_GRAPH));
			Debug.enableDebug(soGetOpt.isActiveOption(OPT_DEBUG));

			Debug.debug("Option set: " + soGetOpt);

			if(soGetOpt.isActiveOption(MARF.WAV))
			{
				MARF.setSampleFormat(MARF.CUSTOM);
				MARF.setSampleLoaderPluginClass(AnyToWAVLoader.class.getName());
			}
			else if(soGetOpt.isActiveOption(MARF.TEXT))
			{
				MARF.setSampleFormat(MARF.TEXT);
				MARF.setSampleLoaderPluginClass(AnyToWAVLoader.class.getName());
				System.out.println("Using text loader");
			}
			else if(soGetOpt.isActiveOption(MARF.TEXT + 1))
			{
				MARF.setSampleFormat(MARF.CUSTOM);
				// XXX:
				MARF.setSampleLoaderPluginClass("XXX");
			}
			else if(soGetOpt.isActiveOption(MARF.TEXT + 2))
			{
				MARF.setSampleFormat(MARF.CUSTOM);
				// XXX:
				MARF.setSampleLoaderPluginClass(RawSampleLoader.class.getName());
			}

			int iMainOption = soGetOpt.getOption(argv[0]);

			switch(iMainOption)
			{
				/*
				 * --------------
				 * Identification
				 * --------------
				 */

				// Single case
				case OPT_IDENT:
				{
					throw new NotImplementedException();
					// ident(getConfigString(argv),
					// soGetOpt.getOption(OPT_DIR_OR_FILE), iExpectedID);
					// ident(getConfigString(argv),
					// soGetOpt.getOption(OPT_SATE_SAMPLE_ARTICLE_FILE),
					// iExpectedID);
					// break;
				}

				// XXX: A directory with files for identification
				case OPT_BATCH_IDENT:
				case MARF.NLP.CLASSIFY:
				{
					// Debug.enableDebug();

					System.out.println("Testing collection size: " + GenericMetaXMLLoader.getInstance().getFileItems().size());

					// Process all the article items for the purpose of the
					// identification
					for(FileItem oFileItem : GenericMetaXMLLoader.getInstance().getFileItems())
					{
						// String strFileID = oFileItem.getFileID();
						String strFileName = oFileItem.getPath();

						// if(new File(strFileID).exists() == false)
						if(new File(strFileName).exists() == false)
						{
							System.err.println("WARNING: File " + strFileName + " does not exist. Ignoring...");
							continue;
							// throw new MARFRuntimeException("File " +
							// strFileName + " does not exist.");
						}

						// Store config and error/successes for that config
						String strConfig = getConfigString(argv);

						// Debug.enableDebug();
						ident(strConfig, oFileItem, iExpectedID);
					}

					Debug.enableDebug(false);

					break;
				}

				/*
				 * --------
				 * Training
				 * --------
				 */

				// Add a single sample to the training set
				case OPT_SINGLE_TRAIN:
				{
					// train(soGetOpt.getOption(OPT_DIR_OR_FILE));
					// System.out.println("Done training with file \"" +
					// soGetOpt.getOption(OPT_DIR_OR_FILE) + "\".");
					// break;
					throw new NotImplementedException();
				}

				// Train on a directory of files
				case OPT_TRAIN:
				case MARF.NLP.TRAIN:
				{
					try
					{
						System.out.println
						(
							"Training collection size: "
							+ GenericMetaXMLLoader.getInstance().getFileItems().size()
						);
						
						soDBPerConfig.syncReferenceData();

						for(FileItem oFileItem : GenericMetaXMLLoader.getInstance().getFileItems())
						{
							// No need to go further for this item if it has no
							// locations defined for training
							if(oFileItem.isLocationless() == true)
							{
								continue;
							}

							String strFilename = oFileItem.getPath();

							if(new File(strFilename).exists() == false)
							{
								throw new MARFRuntimeException("File " + strFilename + " does not exist.");
							}

							// Debug.enableDebug();
							train(oFileItem);
							// Debug.enableDebug(false);
						}
					}
					catch(NullPointerException e)
					{
						// System.err.println("Folder \"" +
						// soGetOpt.getOption(OPT_DIR_OR_FILE) +
						// "\" not found.");
						e.printStackTrace(System.err);
						System.exit(-1);
					}

					// System.out.println("Done training on folder \"" +
					// soGetOpt.getOption(OPT_DIR_OR_FILE) + "\".");

					break;
				}

				/*
				 * -------------------------
				 * Training Set Data Display
				 * -------------------------
				 */
				case OPT_SHOW_TRAINING_SET_DATA:
				{
					System.out.println
					(
						marf.Classification.Classification.loadTrainingSet
						(
							marf.Storage.IStorageManager.DUMP_GZIP_BINARY,
							soGetOpt.getOption(OPT_DIR_OR_FILE)
						)
					);

					break;
				}

				/*
				 * -----
				 * Stats
				 * -----
				 */
				case OPT_STATS:
				{
					soDBPerConfig.restore();
					soDBPerConfig.printStats();
					soDBPerCVE.restore();
					soDBPerCVE.printStats();
					soDBPerCWE.restore();
					soDBPerCWE.printStats();
					break;
				}

				/*
				 * Best Result with Stats
				 */
				case OPT_BEST_SCORE:
				{
					soDBPerConfig.restore();
					soDBPerConfig.printStats(true);
					soDBPerCVE.restore();
					soDBPerCVE.printStats(true);
					soDBPerCWE.restore();
					soDBPerCWE.printStats(true);
					break;
				}

				/*
				 * Reset Stats
				 */
				case OPT_RESET:
				{
					soDBPerConfig.resetStats();
					soDBPerCVE.resetStats();
					soDBPerCWE.resetStats();
					System.out.println("SATE2010App: Statistics has been reset.");
					break;
				}

				/*
				 * Versioning
				 */
				case OPT_VERSION:
				{
					System.out.println("SATE2010 MARF Application, v." + getVersion());
					System.out.println("Using MARF, v." + MARF.getVersion());
					validateVersions();
					break;
				}

				/*
				 * Help
				 */
				case OPT_HELP_SHORT:
				case OPT_HELP_LONG:
				{
					usage();
					break;
				}

				/*
				 * Invalid major option
				 */
				default:
				{
					throw new Exception("Unrecognized option: " + argv[0]);
				}
			} // major option switch

			// Print out SATE 2010 report
			// System.out.println(soReport);
			File oReportFile = new File
			(
				"report-"
				+ getConfigString(argv).replaceAll("\\W", "") + "-"
				+ soGetOpt.getOption(OPT_DIR_OR_FILE) + ".xml"
			);

			Writer oWriter = new FileWriter(oReportFile);
			oWriter.write(soReport.toString());
			oWriter.close();

		} // try

		/*
		 * No arguments have been specified
		 */
		catch(ArrayIndexOutOfBoundsException e)
		{
			System.err.println("No arguments have been specified.");
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);

			usage();
		}

		/*
		 * MARF-specific errors
		 */
		catch(MARFException e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
		}

		/*
		 * Invalid option and/or option argument
		 */
		catch(Exception e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
			usage();
		}

		/*
		 * Regardless whatever happens, close the db connection.
		 */
		finally
		{
			try
			{
				Debug.debug("Closing DB connection...");

				if(soDBPerConfig != null)
				{
					soDBPerConfig.close();
				}

				if(soDBPerCVE != null)
				{
					soDBPerCVE.close();
				}

				if(soDBPerCWE != null)
				{
					soDBPerCWE.close();
				}
			}
			catch(Exception e)
			{
				Debug.debug("Closing DB connection failed: " + e.getMessage());
				e.printStackTrace(System.err);
				System.exit(-1);
			}
		}
	}

	public static final Integer DEFAULT_SEVERITY = 2;
	public static final Double DEFAULT_PROBABILITY = 0.0;

	public static final Double DISTANCE_THRESHOLD = 0.1;
	public static final Double SIMILARITY_THRESHOLD = 0.001;
	public static final Double PROBABILITY_THRESHOLD = 0.9;

	/**
	 * Identifies a subject using MARF given configuration, filename, and
	 * possibly expected subject ID.
	 * 
	 * @param pstrConfig
	 *            configuration string for stats
	 * @param pstrFilename
	 *            name of the file with data sample to identify
	 * @param piExpectedID
	 *            expected subject ID; if -1 then no stats is kept
	 * 
	 * @throws MARFException
	 *             in case of any error while processing is in MARF
	 * @since 0.3.0.5
	 */
	public static final void ident(String pstrConfig, FileItem poFileItem, int piExpectedID)
	throws MARFException
	{
		if(poFileItem.isToIgnore() == true)
		{
			System.out.println("The file item below was indicated to be ignored:");
			System.out.println("================================================");
			System.out.println(poFileItem);
			System.out.println("================================================");
			return;
		}
		
		String strFilename = poFileItem.getPath();
		
//		Boolean bQuickTestOnTrainIgnoreOthers = true;
		Boolean bQuickTestOnTrainIgnoreOthers = false;

		// XXX: can be computed with multiple expected answers
		Double dRecall = 0.0;

		List<Integer> oExpectedIDs = null;
		
		/*
		 * If no expected subject is present on the command line,
		 * attempt to fetch it from the database by filename.
		 */
		if(piExpectedID < 0)
		{											
			/*
			iExpectedID = soDBPerConfig.getIDByFilename(strFilename, false);
			
			if(iExpectedID > 0)
			{
				bQuickTestOnTrainIgnoreOthers = false;
			}
			*/
			
			oExpectedIDs = soDBPerConfig.getIDsByFilename(strFilename, false);
			
			if(oExpectedIDs != null && oExpectedIDs.size() > 0)
			{
				bQuickTestOnTrainIgnoreOthers = false;
			}
		}
		else
		{
			oExpectedIDs = new Vector<Integer>();
			oExpectedIDs.add(piExpectedID);
		}

		// For quick testing on known subjects, don't bother go further
		if(bQuickTestOnTrainIgnoreOthers == true)
		{
			return;
		}
		
		int iIdentifiedID = -1;

		// Second best
		int iSecondClosestID = -1;

		ResultSet oResultSet = null;

		String strClassificationClass = null;
		
		long lTimeStart = System.currentTimeMillis();
		{
			if(soGetOpt.isActiveOption(MARF.NLP.CLASSIFY))
			{
				try
				{
					// Setting of the subject is not really necessary for NLP classification here
					oResultSet = mainNLP(new String[] { oExpectedIDs + "", strFilename });
					strClassificationClass = MaxProbabilityClassifier.class.getName();
				}
				catch(MARFException e)
				{
					if(e.getMessage().toLowerCase().matches(".*mark invalid.*"))
					{
						System.err.println("NLP: Ignoring oversized file in file item: " + poFileItem);
						return;
					}
					else
					{
						System.err.println("NLP: could not handle file item: " + poFileItem);
						System.err.println("NLP: exception class: [" + e.getClass().getName() + "]");
						System.err.println("NLP: exception message: [" + e.getMessage() + "]");
						throw e;
					}
				}
			}
			
			// Traditional way
			else
			{
				MARF.setSampleFile(strFilename);
				MARF.recognize();
				oResultSet = MARF.getResultSet();
				strClassificationClass = MARF.getClassification().getClass().getName();
				iIdentifiedID = MARF.queryResultID();
			}

			System.out.println(oResultSet);
		}
		long lTimeFinish = System.currentTimeMillis();

		// XXX: constify
		int iOneSecInMs  = 1000;
		int iOneMinInMs  = 60 * iOneSecInMs;
		int iOneHourInMs = 60 * iOneMinInMs;
		int iOneDayInMs  = 24 * iOneHourInMs;

		long lDuration = lTimeFinish - lTimeStart;

		long lDays = lDuration / iOneDayInMs;
		long lModTimeLeft = lDuration % iOneDayInMs;

		long lHours = lModTimeLeft / iOneHourInMs;
		lModTimeLeft %= iOneHourInMs;

		long lMinutes = lModTimeLeft / iOneMinInMs;
		lModTimeLeft %= iOneMinInMs;

		long lSeconds = lModTimeLeft / iOneSecInMs;
		lModTimeLeft %= iOneSecInMs;

		long lMilliSeconds = lModTimeLeft;

		StringBuffer oDuration = new StringBuffer()
			.append(lDays + "d:" + lHours + "h:" + lMinutes + "m:" + lSeconds +"s:" + lMilliSeconds +"ms:" + lDuration + "ms")
			;

		if(oResultSet.size() == 0)
		{
			System.out.println("Result set is empty.");
		}
		
		// ---------
		// Output up to this number of non-reporting warnings
		//Integer DEFAULT_REPORT_THRESHOLD = 1;
		//Integer iReportThreshold = 0;

		Boolean bProceedToNextResult = true;

		Boolean bMetaLocationsReported = false;
		
		
		//Result oResult = null;
		
		for(Result oResult: oResultSet.getResultSetSorted())
		{
			if(bProceedToNextResult == false)
			{
				// Useless to continue
				break;
			}
			
			// First guess
			//oResult = oResultSet.getResultSetSorted()[0];
			iIdentifiedID = oResult.getID();

			// Second guess
			Result oSecondResult = null;

			if(oResultSet.size() > 1)
			{
				// Second best
				//iSecondClosestID = oResultSet.getSecondClosestID();
				oSecondResult = oResultSet.getSecondClosestResult();
				iSecondClosestID = oSecondResult.getID();
			}
		
		
			String strShortcutWarnings = "";
			
			if(bQuickTestOnTrainIgnoreOthers == true)
			{
				strShortcutWarnings = "WARNING: bQuickTestOnTrainIgnoreOthers MODE IS ON.";
			}
			
			StringBuffer oOutput = new StringBuffer()
				.append(strShortcutWarnings).append("\n")
				.append("                 File: " + strFilename + "\n")
				.append("              Path ID: " + poFileItem.getFileID() + "\n")
				.append("               Config: " + pstrConfig + "\n")
				.append("      Processing time: " + oDuration + "\n")
				.append("         Subject's ID: " + iIdentifiedID + "\n")
				.append("   Subject identified: " + soDBPerConfig.getName(iIdentifiedID) + "\n")
				//.append("Resultset: " + MARF.getResultSet() + "\n")
				.append("ResultSet: " + oResultSet + "\n")
				.append("FileItem: " + poFileItem + "\n")
				;


			Boolean bCorrectGuess = false;
		
			/*
			 * Only collect stats if we have expected subject
			 */
			if(oExpectedIDs != null && oExpectedIDs.size() > 0)
			//if(iExpectedID > 0)
			//for(Integer iExpectedID: oExpectedIDs)
			{
				Integer iExpectedID = -1;
				
				for(Integer iPossibleExpectedID: oExpectedIDs)
				{	
					bCorrectGuess = (iIdentifiedID == iPossibleExpectedID);
					
					// Assume one of the expected IDs
					iExpectedID = iPossibleExpectedID;
	
					if(bCorrectGuess == true)
					{
						//iExpectedID = iPossibleExpectedID;
						break;
					}
				}
					
				oOutput
					//.append("Expected subject's ID: " + iExpectedID + "\n")
					.append("Expected subject's ID: " + iExpectedID + " (possible: " + oExpectedIDs + ")\n")
					.append("     Expected subject: " + soDBPerConfig.getName(iExpectedID) + "\n")
				;
	
				System.out.println(oOutput);
	
				/*
				System.out.println("Expected subject's ID: " + piExpectedID);
				System.out.println("     Expected subject: " + soDBPerConfig.getName(piExpectedID));
				*/
	
				soDBPerConfig.restore();
				
				if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
				{
					soDBPerCWE.restore();
				}
				else
				{
					soDBPerCVE.restore();
				}
				{
					// 1st match
	//				bCorrectGuess = (iIdentifiedID == iExpectedID); 
					soDBPerConfig.addStats(pstrConfig, bCorrectGuess);
	
					if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
					{
						//soDBPerCWE.addStats(soDBPerConfig.getName(iExpectedID), (iIdentifiedID == iExpectedID));
						soDBPerCWE.addStats(soDBPerConfig.getName(iExpectedID), bCorrectGuess);
					}
					else
					{
						//soDBPerCVE.addStats(soDBPerConfig.getName(iExpectedID), (iIdentifiedID == iExpectedID));
						soDBPerCVE.addStats(soDBPerConfig.getName(iExpectedID), bCorrectGuess);
					}
	
					// 2nd best: must be true if either 1st true or second true (or both :))
					boolean bSecondBest =
						bCorrectGuess
						||
						iSecondClosestID == iExpectedID;
	
					soDBPerConfig.addStats(pstrConfig, bSecondBest, true);
	
					if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
					{
						soDBPerCWE.addStats(soDBPerConfig.getName(iExpectedID), bSecondBest, true);
					}
					else
					{
						soDBPerCVE.addStats(soDBPerConfig.getName(iExpectedID), bSecondBest, true);
					}
				}
				soDBPerConfig.dump();
	
				if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
				{
					soDBPerCWE.dump();
				}
				else
				{
					soDBPerCVE.dump();
				}
			} // expected ID processing
	
			oOutput
				.append("       Second Best ID: " + iSecondClosestID + "\n")
				.append("     Second Best Name: " + soDBPerConfig.getName(iSecondClosestID) + "\n")
				.append("            Date/time: " + new Date() + "\n")
	//			.append("----------------------------8<------------------------------" + "\n")
			;


			// XXX: if threshold is small, do not generate a warning
			Boolean bReportWarning = true;
			
			Integer iSeverity = DEFAULT_SEVERITY;
			Double dProbability = DEFAULT_PROBABILITY;
			
			//if(MARF.getClassification() instanceof Distance)
			//if(strClassificationClass.equals(Distance.class.getName()))
			if(strClassificationClass.matches(".*Distance.*"))
			{
				if(Math.abs(oResult.getOutcome()) > DISTANCE_THRESHOLD && !bCorrectGuess)
				{
					bReportWarning = false;
				}
				else
				{
					dProbability = 1 / oResult.getOutcome();
				}

				oOutput
					.append("Outcome o (classifier-specific): ").append(oResult.getOutcome()).append("\n")
					.append("             Distance threshold: ").append(DISTANCE_THRESHOLD).append("\n")
					.append("           Computed raw P = 1/o: ").append(dProbability).append("\n")
					.append("         Warning to be reported: ").append(bReportWarning).append("\n")
					;
				
				// Lookahead
				if(oSecondResult == null || Math.abs(oSecondResult.getOutcome()) > DISTANCE_THRESHOLD)
				{
					bProceedToNextResult = false;
				}
			}

			//if(MARF.getClassification() instanceof CosineSimilarityMeasure)
			if(strClassificationClass.equals(CosineSimilarityMeasure.class.getName()))
			{
				if(Math.abs(oResult.getOutcome()) < 1 - SIMILARITY_THRESHOLD && !bCorrectGuess)
				{
					bReportWarning = false;
				}
				else
				{
					dProbability = oResult.getOutcome();
				}

				oOutput
					.append("Outcome o (classifier-specific): ").append(oResult.getOutcome()).append("\n")
					.append("     Similarity threshold (1-T): ").append(1 - SIMILARITY_THRESHOLD).append("\n")
					.append("             Computed raw P = o: ").append(dProbability).append("\n")
					.append("         Warning to be reported: ").append(bReportWarning).append("\n")
					;

				// Lookahead
				if(oSecondResult == null || Math.abs(oSecondResult.getOutcome()) < 1 - SIMILARITY_THRESHOLD)
				{
					bProceedToNextResult = false;
				}
			}

			if(strClassificationClass.equals(MaxProbabilityClassifier.class.getName()))
			{
				if(Math.abs(oResult.getOutcome()) > PROBABILITY_THRESHOLD && !bCorrectGuess)
				{
					bReportWarning = false;
				}
				else
				{
					if(oResult.getOutcome() >= 0 && oResult.getOutcome() <= 1)
					{
						// Good, that's the way it should be
						dProbability = oResult.getOutcome();
					}
					else
					{
						/*
						dProbability
							= (oResult.getOutcome() - oResultSet.getMinimumResult().getOutcome())
							/ (oResultSet.getMaximumResult().getOutcome() - oResultSet.getMinimumResult().getOutcome());
						*/
						
						// Kludge, to compensate for weird probabilities, until fixed.
						// The only reason to bother about it
						Double dResult = oResult.getOutcome();
						Double dMinimum = oResultSet.getMinimumResult().getOutcome();
						Double dMaximum = oResultSet.getMaximumResult().getOutcome();
						
						if(dMinimum.isInfinite())
						{
							dMinimum = Double.MIN_VALUE;
						}
						
						if(dMaximum.isInfinite())
						{
							dMaximum = Double.MAX_VALUE;
						}
						
						Double dResultShift = (dResult - dMinimum);
						Double dRangeShift = (dMaximum - dMinimum);

						oOutput
							.append("                        Minimum: ").append(dMinimum).append("\n")
							.append("                        Maximum: ").append(dMaximum).append("\n")
							.append("                   Result shift: ").append(dResultShift).append("\n")
							.append("                    Range shift: ").append(dRangeShift).append("\n")
						;

						dProbability = dResultShift / dRangeShift;
					}
				}
				
				oOutput
					.append("Outcome o (classifier-specific): ").append(oResult.getOutcome()).append("\n")
					.append("          Probability threshold: ").append(PROBABILITY_THRESHOLD).append("\n")
					.append("           Computed raw P range: ").append(dProbability).append("\n")
					.append("         Warning to be reported: ").append(bReportWarning).append("\n")
					;

				// Lookahead
				if(oSecondResult == null || Math.abs(oSecondResult.getOutcome()) > PROBABILITY_THRESHOLD)
				{
					bProceedToNextResult = false;
				}
			}

			// Distance can be negative
			dProbability = Math.abs(dProbability);
			
			// XXX
			if(dProbability > 1 || dProbability.isNaN())
			{
				dProbability = 1.0;
			}
			
			oOutput
				.append("          Computed normalized P: ").append(dProbability).append("\n")
				;
	
			System.out.println(oOutput);
			
			if(bReportWarning == true)
			{
				Warning oWarning = new Warning();
		
				String strCWE = null;
				
				// Primary subject of the classification, can be a CVE or CWE
				String strCXE = soDBPerConfig.getName(iIdentifiedID);
				
				if(soGetOpt.isActiveOption(OPT_SATE_CWEID))
				{
					strCWE = strCXE;
				}
				else
				{
					// XXX: For now assume unique CVE->CWE 1-to-1 mapping
					strCWE = SATEDb.MAP_CVES_TO_CWES.get(strCXE);
				}
				
				Integer iCWE = SATEDb.MAP_CWEIDS_TO_IDS.get(strCWE);
				String strName = SATEDb.MAP_CWEIDS_TO_NAMES.get(strCWE);
				iSeverity = SATEDb.MAP_CWES_TO_SEVERITIES.get(strCWE);
	
				oWarning.setCWE(iCWE);
				oWarning.setName(strName + " (" + strCWE + ")");
				
				oWarning.setMARFID(soReport.size() + 1);
				oWarning.setSATEID(soReport.size() + 1);
				//oWarning.setCorrectness("");
				oWarning.setOutput(oOutput.toString());
				
				//if(MARF.getResultSet().size() > 0)
				if(oResultSet.size() > 0)
				{
					//oWarning.setWeaknessGrade(MARF.getResult().getOutcome());
					oWarning.setWeaknessGrade
					(
						iSeverity,
						dProbability,
						//MARF.getResult().getOutcome()
						oResult.getOutcome()
					);
				}
				
				String strCodeFragment = "[code fragment not available at this time due to signal processing techniques]";
				
				String strExplanation =
					//"Classifier for tool_specific_rank:\n" + MARF.getClassification().getClass().getName() + "\n"
					"Classifier for 'tool_specific_rank':\n" + strClassificationClass + "\n"
					+ "(either minimum distance or maximum similarity)\n\n"
					+ "Severity is computed from the CVE/CWE identified.\n"
					+ "Probability is computed from the distance or similarity.\n\n"
					+ "If line number is 0 we could not reconstruct it back from the waveform, yet.\n"
					+ "If it is non-zero, it is possibly approximate.\n\n"
					+ "See tool-specific output for additional details.\n"
					;
	
				// It is prudent and convenient to report the link to the identified CVE
				if(strCXE.matches("CVE-.*"))
				{
					strExplanation +=
						"\n"
						+ "CVE details: http://web.nvd.nist.gov/view/vuln/detail?vulnId="
						+ strCXE
						+ "\n"
						;
				}
				
				// If there are line numbers available somehow, get them.
				
				ILineNumberHeuristic oLineNumberHeuristic = null;

				Integer iHeuristicLineNumber = 0;
				
				Integer iLinesOfText = 0;
				Integer iWords = 0;
				Integer iBytes = 0;
				
				try
				{
					iLinesOfText = Integer.parseInt(poFileItem.getLines());
					//iLinesOfText = Integer.parseInt(oMainFileItem.getLines());
				}
				catch(Exception e)
				{
					iLinesOfText = 0;
				}
				
				try
				{
					iWords = Integer.parseInt(poFileItem.getWords());
					//iWords = Integer.parseInt(oMainFileItem.getWords());
				}
				catch(Exception e)
				{
					iWords = 0;
				}
				
				try
				{
					iBytes = Integer.parseInt(poFileItem.getBytes());
					//iBytes = Integer.parseInt(poFileItem.getBytes());
				}
				catch(Exception e)
				{
					try
					{
						iBytes = new Long(new File(poFileItem.getPath()).length()).intValue();
					}
					catch(Exception e2)
					{
						iBytes = 0;
					}
				}
				
				if(iLinesOfText == 0 && iBytes == 0)
				{
					oLineNumberHeuristic = new FirstLineNumber();
				}
				else
				{
					oLineNumberHeuristic = new MidLineNumber();
				}
				
				if(iLinesOfText > 0)
				{
					iHeuristicLineNumber = oLineNumberHeuristic.computeLineNumber(iLinesOfText, iWords, iBytes);
				}
				else
				{
					iHeuristicLineNumber = oLineNumberHeuristic.computeLineNumberFromBytes(iBytes);
				}

				soDBPerConfig.restore();
				List<FileItem> oReferenceData = soDBPerConfig.getFileItemsBySubject(strCXE);
				
				System.out.println("oReferenceData: " + oReferenceData);
				
				FileItem oMainFileItem = null;
				
				Integer iReferenceFuzz = 0;
				
				// We are likely testing on the training data. 
				if(oReferenceData.contains(poFileItem) || oReferenceData.size() == 0)
				{
					oMainFileItem = poFileItem;
				}
				
				// Easy case when there only one item to choose from
				else if(oReferenceData.size() == 1)
				{
					oMainFileItem = oReferenceData.get(0);
					
					// XXX: Recompute heuristic line fuzz based on the picked item vs. passed item
					//      compute mid-point of passed vs. mid point of referenced; then
					//      subtract -- the fuzz, apply the fuzz to the line number from
					//      from the referenced
					
					Integer iMidLineParam = new MidLineNumber().computeLineNumber(iLinesOfText);
					
					// XXX: may through an exception
					Integer iMidLineRef = new MidLineNumber().computeLineNumber(Integer.parseInt(oMainFileItem.getLines()));
					
					iReferenceFuzz = iMidLineParam - iMidLineRef;
				}
				
				else
				{
					// XXX
					
					// Stage 1: strictest metrics
					for(FileItem oCurrentItem: oReferenceData)
					{
						if
						(
							oCurrentItem.getLines().equals(poFileItem.getLines())
							&& oCurrentItem.getWords().equals(poFileItem.getWords())
							&& oCurrentItem.getBytes().equals(poFileItem.getBytes())
						)
						{
							oMainFileItem = oCurrentItem;
							break;
						}
					}

					// Stage 2: loosened up: no words
					if(oMainFileItem == null)
					{
						for(FileItem oCurrentItem: oReferenceData)
						{
							if
							(
								oCurrentItem.getLines().equals(poFileItem.getLines())
								&& oCurrentItem.getBytes().equals(poFileItem.getBytes())
							)
							{
								oMainFileItem = oCurrentItem;
								break;
							}
						}
					}

					// Stage 3: loosened up: no words no bytes
					if(oMainFileItem == null)
					{
						for(FileItem oCurrentItem: oReferenceData)
						{
							if
							(
								oCurrentItem.getLines().equals(poFileItem.getLines())
							)
							{
								oMainFileItem = oCurrentItem;
								break;
							}
						}
					}

					// Stage 4: loosened up: lines +- 10 within
					if(oMainFileItem == null)
					{
						for(FileItem oCurrentItem: oReferenceData)
						{
							if
							(
								//XXX: May throw exception
								iLinesOfText <= Integer.parseInt(oCurrentItem.getLines()) + 10
								&& iLinesOfText >= Integer.parseInt(oCurrentItem.getLines()) - 10
							)
							{
								oMainFileItem = oCurrentItem;
								break;
							}
						}
					}
					
					// Stage 5: loosened up: bytes +- 1000 within
					if(oMainFileItem == null)
					{
						for(FileItem oCurrentItem: oReferenceData)
						{
							if
							(
								//XXX: May throw exception
								iBytes <= Integer.parseInt(oCurrentItem.getBytes()) + 1000
								&& iBytes >= Integer.parseInt(oCurrentItem.getBytes()) - 1000
							)
							{
								oMainFileItem = oCurrentItem;
								break;
							}
						}
					}

					// Stage 6: pseudo random
					if(oMainFileItem == null)
					{
						oReferenceData.get(new Random().nextInt(oReferenceData.size()));
					}
				}
				
				// Worst case, just take main item to be the parameter if nothing
				// got picked
				if(oMainFileItem == null)
				{
					oMainFileItem = poFileItem;
				}
				
				//Vector<MetaLocation> oMetaLocations = poFileItem.getLocations();
				Vector<MetaLocation> oMetaLocations = oMainFileItem.getLocations();

				Integer iFragmentLines = 1;
				
				//oLineNumberHeuristic.computeLineNumber()
				// Line cannot be found
				if(oMetaLocations.size() == 0)
				{
					if(iHeuristicLineNumber > 0)
					{
						strCodeFragment = new LineContextReader(strFilename, iHeuristicLineNumber).getFragmentWithContext();
					}
					
					oWarning.addLocation
					(
						//0,
						iHeuristicLineNumber,
						strFilename,
						strCodeFragment,
						strExplanation,
						poFileItem.getFileID()
					);
				}
				
				// Report lines for precisely known files
				else
				{
					// To avoid listing explanations from meta locations more than
					// once.
					ArrayList<String> oAppearedExplanations = new ArrayList<String>();
					
					// Do not report the meta locations more than once per file
					// if more than one category matches it
					//if(bMetaLocationsReported == false)
					{
						for(MetaLocation oLocation: oMetaLocations)
						{
							if(oLocation.isToIgnore() == true)
							{
								System.out.println("The meta location below was indicated to be ignored:");
								System.out.println("====================================================");
								System.out.println(oLocation);
								System.out.println("====================================================");
								continue;
							}
							
							// Verify the location belongs to the correct category if it needs to be reported
							if(oLocation.containsSubject(strCXE) == false)
							{
								continue;
							}
							
							Integer iLineNumber = 0;
							
							try
							{
								iLineNumber = Integer.parseInt(oLocation.getLineNumber());
							}
							catch(NumberFormatException e)
							{
								//iLineNumber = 0;
								iLineNumber = iHeuristicLineNumber;
							}
							
							if(iLineNumber == null || iLineNumber <= 0)
							{
								//iLineNumber = 0;
								iLineNumber = iHeuristicLineNumber;
							}
							
							// Has effect in testing files that either shrank or grew
							iLineNumber += iReferenceFuzz;

							try
							{
								iFragmentLines = Integer.parseInt(oLocation.getFragmentLines());
							}
							catch(Exception e)
							{
								iFragmentLines = 1;
							}
							
							// For fuzzy test matches just in case include more of the fuzzy region
							// for visibility
							iFragmentLines += Math.abs(iReferenceFuzz);
							
							if(iLineNumber > 0)
							{
								strCodeFragment	= new LineContextReader(strFilename, iLineNumber, iFragmentLines)
									.getFragmentWithContext();
							}
							
							if
							(
								oLocation.getExplanation().equals("") == false
								&& oLocation.getExplanation().equals(MetaLocation.DEFAULT_EXPLANATION) == false
							)
							{
								if(oAppearedExplanations.contains(oLocation.getExplanation()))
								{
									// skip adding
								}
								else
								{
									oAppearedExplanations.add(oLocation.getExplanation());
									
									strExplanation +=
										"Possible explanation material:\n"
										+ "=====================================\n"
										+ "Type: " + oLocation.getType() + "\n"
										+ "CxE: [" + strCXE + "]\n"
										+ (strCXE.matches(".*CWE.*") ? "" : oLocation.getExplanation()) + "\n"
										+ "=====================================\n"
										;
								}
							}
							
							oWarning.addLocation
							(
								iLineNumber,
								strFilename,
								strCodeFragment,
								strExplanation,
								poFileItem.getFileID()
							);
						} // all meta locations
						
						bMetaLocationsReported = true;
					}
				}
				
				soReport.addWarning(oWarning);
			} // if report a warning
			else
			{
				// Useless to continue further cycling through the results.
				break;
			}
		} // for each result in the result set
	}

	/**
	 * Updates training set with a new sample from a given file.
	 * 
	 * @param pstrFilename
	 *            name of the wave file with voice sample train the system on
	 * 
	 * @throws MARFException
	 *             in case of any error while processing is in MARF
	 */
	public static final void train(FileItem poFileItem) throws MARFException
	{
		if(poFileItem.isToIgnore() == true)
		{
			System.out.println("The file item below was indicated to be ignored:");
			System.out.println("================================================");
			System.out.println(poFileItem);
			System.out.println("================================================");
			return;
		}

		String strFilename = poFileItem.getPath();

		MARF.setSampleFile(strFilename);

		// int iID = soDBPerConfig.getIDByFilename(strFilename, true);
		List<Integer> oIDs = soDBPerConfig.getIDsByFilename(strFilename, true);

		// Debug.debug("Training on ID: " + iID + ", and FileItem [" +
		// poFileItem + "]");
		Debug.debug("Training on IDs: " + oIDs + ", and FileItem [" + poFileItem + "]");

		// if(iID == -1)
		if(oIDs == null || oIDs.size() == 0)
		{
			// System.out.println("No subject found for \"" + strFilename +
			// "\" for training.");
			System.out.println("No subject(s) found for \"" + strFilename + "\" for training.");
		}
		else
		{
			if(soGetOpt.isActiveOption(MARF.NLP.TRAIN))
			{
				for(Integer iID : oIDs)
				{
					mainNLP(new String[] { iID + "", strFilename });
				}
			}
			else
			{
				for(Integer iID : oIDs)
				{
					MARF.setCurrentSubject(iID);
					MARF.train();
				}
			}
		}
	}

	/**
	 * Displays application's usage information and exits.
	 */
	private static final void usage()
	{
		System.out.println("Usage:\n"
				+ "  java MARFCATApp      --train <samples-dir> [options]        -- train mode\n"
				+ "                       --single-train <sample> [options]      -- add a single sample to the training set\n"
				+ "                       --ident <sample> [options]             -- identification mode\n"
				+ "                       --batch-ident <samples-dir> [options]  -- batch identification mode\n"
				+ "                       --train-nlp [ --debug ] [ OPTIONS ] <language> <corpus-file>\n"
				+ "                       --ident-nlp [ --debug ] [ OPTIONS ] foo <bar|corpus-file>\n\n"
				+ "                       --debug                                -- include verbose debug output\n"
				+ "                       --gui                                  -- use GUI as a user interface\n"
				+ "                       --stats=[per-config|per-writer|both]   -- display stats (default is per-config)\n"
				+ "                       --best-score                           -- display best classification result\n"
				+ "                       --reset                                -- reset stats\n"
				+ "                       --version                              -- display version info\n"
				+ "                       --help | -h                            -- display this help and exit\n\n" +

				"Options (one or more of the following):\n\n" +

				"Loaders:\n\n"
				+ "  -wav          - assume WAVE files loading (default)\n"
				+ "  -text         - assume loading of text samples\n"
				+ "  -tiff         - assume loading of TIFF samples\n"
				+ "\n" +

				"Preprocessing:\n\n"
				+ "  -silence      - remove silence (can be combined with any of the below)\n"
				+ "  -noise        - remove noise (can be combined with any of the below)\n"
				+ "  -raw          - no preprocessing\n"
				+ "  -norm         - use just normalization, no filtering\n"
				+ "  -low          - use low-pass FFT filter\n"
				+ "  -high         - use high-pass FFT filter\n"
				+ "  -boost        - use high-frequency-boost FFT preprocessor\n"
				+ "  -band         - use band-pass FFT filter\n"
				+ "  -bandstop     - use band-stop FFT filter\n"
				+ "  -endp         - use endpointing\n"
				+ "  -lowcfe       - use low-pass CFE filter\n"
				+ "  -highcfe      - use high-pass CFE filter\n"
				+ "  -bandcfe      - use band-pass CFE filter\n"
				+ "  -bandstopcfe  - use band-stop CFE filter\n"
				+ "\n" +

				"Feature Extraction:\n\n"
				+ "  -lpc          - use LPC\n"
				+ "  -fft          - use FFT\n"
				+ "  -minmax       - use Min/Max Amplitudes\n"
				+ "  -randfe       - use random feature extraction\n"
				+ "  -aggr         - use aggregated FFT+LPC feature extraction\n"
				+ "  -f0           - use F0 (pitch, or fundamental frequency; NOT IMPLEMENTED\n"
				+ "  -segm         - use Segmentation (NOT IMPLEMENTED)\n"
				+ "  -cepstral     - use Cepstral analysis (NOT IMPLEMENTED)\n"
				+ "\n" +

				"Classification:\n\n"
				+ "  -nn           - use Neural Network\n"
				+ "  -cheb         - use Chebyshev Distance\n"
				+ "  -eucl         - use Euclidean Distance\n"
				+ "  -mink         - use Minkowski Distance\n"
				+ "  -diff         - use Diff-Distance\n"
				+ "  -zipf         - use Zipf's Law-based classifier\n"
				+ "  -randcl       - use random classification\n"
				+ "  -markov       - use Hidden Markov Models (NOT IMPLEMENTED)\n"
				+ "  -hamming      - use Hamming Distance\n"
				+ "  -cos          - use Cosine Similarity Measure\n"
				+ "\n" +

				"NLP/Ngrams/Smoothing:\n\n"
				+ "  -interactive  - interactive mode for classification instead of reading from a file\n"
				+ "  -char         - use characters as n-grams (should always be present for this app)\n\n"
				+ "  -unigram      - use UNIGRAM model\n"
				+ "  -bigram       - use BIGRAM model\n"
				+ "  -trigram      - use TRIGRAM model\n\n"
				+ "  -mle          - use MLE\n"
				+ "  -add-one      - use Add-One smoothing\n"
				+ "  -add-delta    - use Add-Delta (ELE, d=0.5) smoothing\n"
				+ "  -witten-bell  - use Witten-Bell smoothing\n"
				+ "  -good-turing  - use Good-Turing smoothing\n" +

				"\n"
				+ "  -spectrogram  - dump spectrogram image after feature extraction\n"
				+ "  -graph        - dump wave graph before preprocessing and after feature extraction\n"
				+ "  <integer>     - expected subject ID\n"
				+ "\n");

		System.exit(0);
	}

	/**
	 * Retrieves String representation of the application's version.
	 * 
	 * @return version String
	 */
	public static final String getVersion()
	{
		return MAJOR_VERSION + "." + MINOR_VERSION + "." + REVISION;
	}

	/**
	 * Retrieves integer representation of the application's version.
	 * 
	 * @return integer version
	 */
	public static final int getIntVersion()
	{
		return MAJOR_VERSION * 100 + MINOR_VERSION * 10 + REVISION;
	}

	/**
	 * Makes sure the applications isn't run against older MARF version. Exits
	 * with 1 if the MARF version is too old.
	 */
	public static final void validateVersions()
	{
		if(MARF.getDoubleVersion() < (0 * 100 + 3 * 10 + 0 + .6))
		{
			System.err.println("Your MARF version (" + MARF.getVersion()
					+ ") is too old. This application requires 0.3.0.6 or above.");

			System.exit(1);
		}
	}

	/**
	 * Composes the current configuration of in a string form.
	 * 
	 * @param pstrArgv
	 *            set of configuration options passed through the command line;
	 *            can be null or empty. If latter is the case, MARF itself is
	 *            queried for its numerical set up inside.
	 * 
	 * @return the current configuration setup
	 */
	public static final String getConfigString(String[] pstrArgv)
	{
		// Store config and error/successes for that config
		String strConfig = "";

		if(pstrArgv != null && pstrArgv.length > 2)
		{
			// Get config from the command line
			for(int i = 2; i < pstrArgv.length; i++)
			{
				// We do NOT want numerical IDs
				try
				{
					Integer.parseInt(pstrArgv[i]);
				}
				catch(NumberFormatException e)
				{
					// Skip .xml file names
					if(pstrArgv[i].matches(".*\\.xml"))
					{
						continue;
					}

					strConfig += pstrArgv[i] + " ";
				}
			}
		}
		else
		{
			// Query MARF for it's current config
			strConfig = MARF.getConfig();
		}

		return strConfig;
	}

	/**
	 * Sets default MARF configuration parameters as normalization for
	 * preprocessing, FFT for feature extraction, Euclidean distance for
	 * training and classification with no spectrogram dumps and no debug
	 * information, assuming WAVE file format.
	 * 
	 * @throws MARFException
	 */
	public static final void setDefaultConfig() throws MARFException
	{
		/*
		 * Default MARF setup
		 */
		MARF.setPreprocessingMethod(MARF.DUMMY);
		MARF.setFeatureExtractionMethod(MARF.FFT);
		MARF.setClassificationMethod(MARF.EUCLIDEAN_DISTANCE);
		MARF.setDumpSpectrogram(false);
		// MARF.setSampleFormat(MARF.WAV);
		// MARF.setSampleFormat(MARF.TEXT);
		MARF.setSampleFormat(MARF.CUSTOM);
		MARF.setSampleLoaderPluginClass(AnyToWAVLoader.class.getName());

		Debug.enableDebug(false);
	}

	/**
	 * Customizes MARF's configuration based on the options.
	 * 
	 * @throws MARFException
	 *             if some options are out of range
	 */
	public static final void setCustomConfig() throws MARFException
	{
		ModuleParams oParams = new ModuleParams();

		for(int iPreprocessingMethod = MARF.MIN_PREPROCESSING_METHOD; iPreprocessingMethod <= MARF.MAX_PREPROCESSING_METHOD; iPreprocessingMethod++)
		{
			if(soGetOpt.isActiveOption(iPreprocessingMethod))
			{
				MARF.setPreprocessingMethod(iPreprocessingMethod);

				switch(iPreprocessingMethod)
				{
					// Endpointing did not respond well to silence removal
					// and the definition of RAW assumes none of that is done.
					// Endpointing also has some extra params of its own.
					case MARF.RAW:
					case MARF.ENDPOINT:

					case MARF.DUMMY:
					case MARF.HIGH_FREQUENCY_BOOST_FFT_FILTER:
					case MARF.HIGH_PASS_FFT_FILTER:
					case MARF.LOW_PASS_FFT_FILTER:
					case MARF.BANDPASS_FFT_FILTER:
					case MARF.BAND_STOP_FFT_FILTER:
					case MARF.HIGH_PASS_BOOST_FILTER:
					case MARF.BAND_PASS_CFE_FILTER:
					case MARF.BAND_STOP_CFE_FILTER:
					case MARF.LOW_PASS_CFE_FILTER:
					case MARF.HIGH_PASS_CFE_FILTER:
					{
						// Kludge
						if(soGetOpt.isActiveOption(OPT_SILENCE_NOISE_REMOVAL))
						{
							oParams.addPreprocessingParam(true);
							oParams.addPreprocessingParam(true);
						}
						else
						{
							// Normalization and filters seem to respond better
							// to silence removal.
							// The setting of the third protocol parameter
							// (silence threshold)
							// is yet to be implemented here.
							oParams.addPreprocessingParam(soGetOpt.isActiveOption(OPT_NOISE_REMOVAL));
							oParams.addPreprocessingParam(soGetOpt.isActiveOption(OPT_SILENCE_REMOVAL));
						}

						break;
					}

					default:
					{
						assert false : "Not implemented valid preprocessing configuration parameter: " + iPreprocessingMethod;
					}
				} // switch

				break;
			}
		}

		for(int iFeatureExtractionMethod = MARF.MIN_FEATUREEXTRACTION_METHOD; iFeatureExtractionMethod <= MARF.MAX_FEATUREEXTRACTION_METHOD; iFeatureExtractionMethod++)
		{
			if(soGetOpt.isActiveOption(iFeatureExtractionMethod))
			{
				MARF.setFeatureExtractionMethod(iFeatureExtractionMethod);

				switch(iFeatureExtractionMethod)
				{
					case MARF.FFT:
					case MARF.LPC:
					case MARF.RANDOM_FEATURE_EXTRACTION:
					case MARF.MIN_MAX_AMPLITUDES:
					case MARF.F0:
					case MARF.CEPSTRAL:
					case MARF.SEGMENTATION:
						// For now do nothing; customize when these methods
						// become parameterizable.
						break;

					case MARF.FEATURE_EXTRACTION_AGGREGATOR:
					{
						// For now aggregate FFT followed by LPC until
						// it becomes customizable
						oParams.addFeatureExtractionParam(new Integer(MARF.FFT));
						oParams.addFeatureExtractionParam(null);
						oParams.addFeatureExtractionParam(new Integer(MARF.LPC));
						oParams.addFeatureExtractionParam(null);
						break;
					}

					default:
						assert false;
				} // switch

				break;
			}
		}

		for(int iClassificationMethod = MARF.MIN_CLASSIFICATION_METHOD; iClassificationMethod <= MARF.MAX_CLASSIFICATION_METHOD; iClassificationMethod++)
		{
			if(soGetOpt.isActiveOption(iClassificationMethod))
			{
				MARF.setClassificationMethod(iClassificationMethod);

				switch(iClassificationMethod)
				{
					case MARF.NEURAL_NETWORK:
					{
						// Dump/Restore Format of the TrainingSet
						oParams.addClassificationParam(TrainingSet.DUMP_GZIP_BINARY);

						// Training Constant
						oParams.addClassificationParam(0.5);

						// Epoch number
						oParams.addClassificationParam(20);

						// Min. error
						oParams.addClassificationParam(0.1);

						break;
					}

					case MARF.HAMMING_DISTANCE:
					{
						// Dump/Restore Format
						oParams.addClassificationParam(TrainingSet.DUMP_GZIP_BINARY);

						// Type of hamming comparison
						oParams.addClassificationParam(HammingDistance.STRICT_DOUBLE);

						break;
					}

					case MARF.MINKOWSKI_DISTANCE:
					{
						// Dump/Restore Format
						oParams.addClassificationParam(TrainingSet.DUMP_GZIP_BINARY);

						// Minkowski Factor
						oParams.addClassificationParam(6.0);

						break;
					}

					case MARF.EUCLIDEAN_DISTANCE:
					case MARF.CHEBYSHEV_DISTANCE:
					case MARF.MAHALANOBIS_DISTANCE:
					case MARF.RANDOM_CLASSIFICATION:
					case MARF.DIFF_DISTANCE:
					case MARF.MARKOV:
					case MARF.ZIPFS_LAW:
					case MARF.COSINE_SIMILARITY_MEASURE:
					{
						// Dump/Restore Format
						oParams.addClassificationParam(TrainingSet.DUMP_GZIP_BINARY);

						// For now do nothing; customize when these methods
						// become parameterizable.
						break;
					}

					default:
						assert false : "Unrecognized classification module";
				} // switch

				// Method is found, break out of the look up loop
				break;
			}
		}

		// Assign meaningful params only
		if(oParams.size() > 0)
		{
			MARF.setModuleParams(oParams);
		}
	}

	/**
	 * The main.
	 * 
	 * @param argv
	 *            command-line arguments
	 * @throws MARFException 
	 */
	// public static ArrayList<Integer> mainNLP(String[] argv)
	public static ResultSet mainNLP(String[] argv)
	throws MARFException
	{
		// ArrayList<Integer> oResults = new ArrayList<Integer>();

		try
		{
			System.out.println("\n" + "SATE2010 NLP Application, $Revision: 1.33 $, $Date: 2011/01/14 00:31:30 $\n"
					+ "Serguei A. Mokhov, mokhov@cse.concordia.ca\n" + "August 2010\n");

			//int iValidOptions = soGetOpt.getValidOptions().size();

			String strFilename = argv[1];
			String strSubject = argv[0];

			System.out.println("Filename: [" + strFilename + "]");
			System.out.println("Subject: [" + strSubject + "]");

			MARF.NLP.setLanguage(strSubject);

			String strSentence = "";
			Reader oReader = null;

			if(soGetOpt.isActiveOption(MARF.NLP.TRAIN) || !soGetOpt.isActiveOption(MARF.NLP.INTERACTIVE))
			{
				FileInputStream oFIS = new FileInputStream(strFilename);
				oReader = new BufferedReader(new InputStreamReader(oFIS));
			}
			else
			// interactive
			{
				System.out.println("\nEntering interactive mode... Type \\q to exit.");

				System.out.print("sentence> ");

				strSentence = getLine();

				if(strSentence == null || strSentence.startsWith("\\q"))
				{
					System.exit(0);
				}

				oReader = new StringReader(strSentence);
			}

			// XXX: constify
			// Integer MAX_READAHEAD_MARK = 65536;
			// Integer MAX_READAHEAD_MARK = 131072;
			// Integer MAX_READAHEAD_MARK = 262144;
			// Max length file wireshark-1.2.0/plugins/giop/packet-parlay.c is 3164610
			// Integer MAX_READAHEAD_MARK = 3200000;
			// Integer MAX_READAHEAD_MARK = 3383210;
			// Integer MAX_READAHEAD_MARK = 4000000;
			Integer MAX_READAHEAD_MARK = 6000000;

			soTokenizer = new NLPStreamTokenizer(oReader, MAX_READAHEAD_MARK);

			// Set up the tokenizer
			setupStreamTokenizer();

			for(Serializable oKey : soGetOpt.getActiveOptions().keySet())
			{
				// int iOption = soGetOpt.getOption(argv[i]);
				int iOption = soGetOpt.getOption(oKey.toString());
				// Set n-gram model options
				// for(int i = 0; i < iValidOptions; i++)
				// {
				// int iOption = soGetOpt.getOption(argv[i]);

				switch(iOption)
				{
					case MARF.ENgramModels.UNIGRAM:
						MARF.NLP.setNgramModel(MARF.ENgramModels.UNIGRAM);
						System.out.println("UNIGRAM model has been selected.");
						break;

					case MARF.ENgramModels.BIGRAM:
						MARF.NLP.setNgramModel(MARF.ENgramModels.BIGRAM);
						System.out.println("BIGRAM model has been selected.");
						break;

					case MARF.ENgramModels.TRIGRAM:
						MARF.NLP.setNgramModel(MARF.ENgramModels.TRIGRAM);
						System.out.println("TRIGRAM model has been selected.");
						break;
				}
			}

			// Determine statistical smoothing method
			StatisticalEstimator oStatisticalEstimator = null;

			for(Serializable oKey : soGetOpt.getActiveOptions().keySet())

			// for(int i = 0; i < iValidOptions; i++)
			{
				// int iOption = soGetOpt.getOption(argv[i]);
				int iOption = soGetOpt.getOption(oKey.toString());

				// Debug.debug("Attempting " + argv[i] +
				// " as a Stat's estimator (" + iOption + ")");

				switch(iOption)
				{
					// Dump raw rank and frequency
					case MARF.NLP.RAW_ZIPFS_LAW_DUMP:
					{
						// bDumpLog = false;
						break;
					}

					case MARF.EStatisticalEstimators.MLE:
					{
						oStatisticalEstimator = new MLE();
						System.out.println("MLE Statistical Estimator has been selected.");
						break;
					}

					case MARF.EStatisticalEstimators.ADD_ONE:
					{
						oStatisticalEstimator = new AddOne();
						System.out.println("Add-One Statistical Estimator has been selected.");
						break;
					}

					case MARF.EStatisticalEstimators.ADD_DELTA:
					case MARF.EStatisticalEstimators.ELE:
					{
						// 0.5 = ELE
						oStatisticalEstimator = new AddDelta(0.5);
						System.out.println("ELE Statistical Estimator has been selected.");
						break;
					}

					case MARF.EStatisticalEstimators.WITTEN_BELL:
					{
						oStatisticalEstimator = new WittenBell();
						System.out.println("Witten-Bell Statistical Estimator has been selected.");
						break;
					}

					case MARF.EStatisticalEstimators.GOOD_TURING:
					{
						oStatisticalEstimator = new GoodTuring();
						System.out.println("Good-Turing Statistical Estimator has been selected.");
						break;
					}

					case MARF.EStatisticalEstimators.SLI:
					{
						oStatisticalEstimator = new SLI();
						System.out.println("SLI Statistical Estimator is not implemented; only n-gram counts will be used.");
						break;
					}

					case MARF.EStatisticalEstimators.GLI:
					{
						oStatisticalEstimator = new GLI();
						System.out.println("GLI Statistical Estimator is not implemented; only n-gram counts will be used.");
						break;
					}

					case MARF.EStatisticalEstimators.KATZ_BACKOFF:
					{
						oStatisticalEstimator = new KatzBackoff();
						System.out
								.println("Katz Backoff Statistical Estimator is not implemented; only n-gram counts will be used.");
						break;
					}

					default:
					{
						Debug.debug("Option " + soGetOpt.getOption(oKey.toString()) + " ignored.");
					}
				}

				// Debug.debug("estimator so far: " + oStatisticalEstimator);

				// To differentiate written files from defaults,
				// add options as suffixes
				// strFilename += argv[i];
				strFilename += oKey;
			}

			if(oStatisticalEstimator == null)
			{
				throw new NullPointerException("No Statistical Estimator has been assigned. Check your options!");
			}

			oStatisticalEstimator.setStreamTokenizer(soTokenizer);

			if(soGetOpt.isActiveOption(MARF.NLP.ZIPFS_LAW_CHEAT))
			{
				ZipfLaw oZipfLawStats = new ZipfLaw(strFilename);
				oZipfLawStats.collectStatistics(soTokenizer);

				// Flush the results
				oZipfLawStats.dumpAll(); // Goes to STDOUT
				oZipfLawStats.dumpGraphValues(); // Goes to a file <file>.csv
			}

			MaxProbabilityClassifier oMaxProbabilityClassifier = new MaxProbabilityClassifier(oStatisticalEstimator);

			if(soGetOpt.isActiveOption(MARF.NLP.TRAIN))
			{
				oMaxProbabilityClassifier.train();
			}
			else
			{
				oMaxProbabilityClassifier.classify();

				System.out.println("\nSubject identified: [" + oMaxProbabilityClassifier.getResult().getDescription() + "]");

				// oResults.add(oMaxProbabilityClassifier.getResult().getID());
				// oResults.add(oMaxProbabilityClassifier.getResultSet().
				// getSecondClosestID());
			}

			if(soGetOpt.isActiveOption(MARF.NLP.INTERACTIVE))
			{
				System.out.println("\nEntering interactive mode... Type \\q to exit.");

				while(true)
				{
					System.out.print("sentence> ");
					strSentence = getLine();

					if(strSentence == null || strSentence.startsWith("\\q"))
					{
						break;
					}

					oReader = new StringReader(strSentence);
					soTokenizer = new NLPStreamTokenizer(oReader);
					setupStreamTokenizer();

					oStatisticalEstimator.setStreamTokenizer(soTokenizer);

					oMaxProbabilityClassifier = new MaxProbabilityClassifier(oStatisticalEstimator);
					oMaxProbabilityClassifier.classify();

					System.out.println("Subject identified: [" + oMaxProbabilityClassifier.getResult().getDescription() + "]");
				}
			}

			// Make sure to release any stream resources for the cases
			// of the repeated invocations within the same JVM and slow
			// garbage collection.
			oReader.close();
			oReader = null;
			
			return oMaxProbabilityClassifier.getResultSet();
		}
		catch(MARFException e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
			throw e;
		}
		catch(Exception e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
			throw new MARFException(e);
		}
	}

	/**
	 * Sets default stream tokenizer.
	 */
	private static final void setDefaultStreamTokenizer()
	{
		/*
		 * Default setup, which will be customized below if any of the options
		 * are supplied
		 */

		// At the beginning make everything other than
		// letters of the English alphabet blanks
		// soTokenizer.whitespaceChars('z' + 1, '\u00FF');
		// soTokenizer.whitespaceChars('\0', 'A' - 1);
		// Fold everything to the lower case
		// soTokenizer.lowerCaseMode(true);
		// Never care about EOLs
		// soTokenizer.eolIsSignificant(false);
	}

	/**
	 * Sets customized stream tokenizer based on options.
	 */
	private static final void setupStreamTokenizer()
	{
		// Set up the tokenizer
		setDefaultStreamTokenizer();

		// Make every character a token
		// if(soGetOpt.isActiveOption(MARF.NLP.CHARACTER_MODE))
		{
			soTokenizer.resetSyntax();
			soTokenizer.eolIsSignificant(true);
			// soTokenizer.eolIsSignificant(false);
			// soTokenizer.lowerCaseMode(true);
		}

		// Make case-sensitive
		// if(soGetOpt.isActiveOption(MARF.NLP.CASE_SENSITIVE))
		{
			soTokenizer.lowerCaseMode(false);
		}

		// Parse numbers as well
		// if(soGetOpt.isActiveOption(MARF.NLP.PARSE_NUMBERS))
		{
			soTokenizer.parseNumbers();
		}

		// Make "'" a word char, so that things like "don't", "won't"
		// "Zipf's" are not split into two parts
		soTokenizer.wordChars('\'', '\'');

		// Treat "some string" as a token
		// if(soGetOpt.isActiveOption(MARF.NLP.PARSE_QUOTED_STRINGS))
		{
			soTokenizer.quoteChar('"');
		}

		// Make "?", ".", and "!" special, even if it says ordinary
		// if(soGetOpt.isActiveOption(MARF.NLP.PARSE_ENDS_OF_SENTENCE))
		{
			soTokenizer.ordinaryChar('?');
			soTokenizer.ordinaryChar('.');
			soTokenizer.ordinaryChar('!');
		}
	}

	/**
	 * Reads off a line from STDIN.
	 * 
	 * @return a line string terminated by an end of line
	 */
	public static final String getLine()
	{
		try
		{
			if(soReader == null)
			{
				soReader = new BufferedReader(new InputStreamReader(System.in));
			}

			return soReader.readLine();
		}
		catch(IOException e)
		{
			System.err.println(e.getMessage());
			e.printStackTrace(System.err);
			// Debug.elog(e);
			System.exit(1);
		}

		assert false;
		return "[we have reached a complete nonsense]";
	}
}

// EOF
